type
 Int8 = System.ShortInt;
 Int16 = System.SmallInt;
 Int32 = System.Longint;
{ Int64 = System.Comp;}

 UInt8 = System.Byte;
 UInt16 = System.Word;
 UInt32 = Cardinal;

 Bool8 = System.Boolean;
 Bool16 = System.WordBool;
 Bool32 = System.LongBool;

type
 TCPUVendor=(venAMD, venCentaur, venCyrix, venIntel, venNexGen, venRise, venUMC, venTransMeta, venUnknown);
 TVendorIdString=String[12];
 TCPUType=(typePrimary, typeOverDrive, typeSecondary, typeReserved);
const
 CPUVendors:array[low(TCPUVendor)..high(TCPUVendor)] of ansistring=(
  'AMD','Centaur/IDT','Cyrix/VIA','Intel','NexGen','Rise','Transmeta','UMC','Unknown');
 VendorIdStrings:array[low(TCPUVendor)..high(TCPUVendor)] of TVendorIdString=(
  'AuthenticAMD', 'CentaurHauls', 'CyrixInstead',
  'GenuineIntel', 'NexGenDriven', 'RiseRiseRise',
  'GenuineTMx86', ' UMC UMC UMC', 'Other Vendor');
 CPUTypes:array[low(TCPUType)..high(TCPUType)] of ansistring=(
  'Primary processor',
  'OverDrive processor',
  'Dual processor (MP)',
  'Reserved');

type
 TCPUFamily=UInt8;
 TCPUModel=UInt8;
 TCPUStepping=UInt8;
 TCPUPSN=array[0..2] of UInt32;

var
 maxLevel:UInt32;
 baseLevel:UInt32;
 vendor:TCPUVendor;
 vendorIdString:TVendorIdString;
 cpuType:TCPUType;
 family:TCPUFamily;
 model:TCPUModel;
 stepping:TCPUStepping;
 psn:TCPUPSN;
 clockrate:integer;
 timer_rate:Int64;

var
 resi:ansistring;
 critical:trtlCRITICALSECTION;
 sn:string;

function GetCPUClockrate(family:UInt8):integer; forward;

function CPUidSupported:Bool8; assembler; {&USES EAX, ECX}
asm
 pushfd
 pushfd               { Get original EFLAGS}
 pop    eax
 mov    ecx, eax
 xor    eax, 200000h { Flip id bit in eflags }
 push   eax           { Save new eflags value on stack }
 popfd                { Replace current eflags value }
 pushfd               { Get new eflags }
 pop    eax           { Store new eflags in eax }
 popfd                { Restore original eflags }
 xor    eax, ecx      { Can not toggle ID bit}
 setnz  al
end;

function Is_386:Bool8; assembler; {&USES EAX, ECX, EDX}
asm
 mov dx, sp         { save current stack pointer }
 and sp, $fffc      { align stack to avoid AC fault }
 pushfd
 pop eax
 mov ecx, eax
 xor eax, $40000    { xor AC bit in EFLAGS }
 push eax
 popfd
 pushfd
 pop eax
 xor eax, ecx       { Is AC toggled ? }
 setz al            { if not, we have a 386 }
 and sp, $fffc      { align stack to prevent AC fault }
 push ecx
 popfd              { restore original AC bit }
 mov sp, dx         { restore original stack pointer }
end;

function Is_486:Bool8; assembler; {&USES EAX, ECX}
asm
 pushf                  { Get original EFLAGS}
 pop eax
 mov ecx, eax           { save EFLAGS }
 xor eax, $200000       { Flip ID bit in EFLAGS}
 push eax               { Save new EFLAGS value on stack}
 popf                   { Replace current EFLAGS value}
 pushf                  { Get new EFLAGS}
 pop eax                { Store new EFLAGS in EAX}
 xor eax, ecx           { Toggle ID bit}
 setz al
 push ecx              { restore old EFLAGS }
 popf
end;

procedure callCPUid(input:UInt32; var regeax, regebx, regecx, regedx:UInt32); assembler;pascal;
asm
 pushad
 mov eax, input
 db 0fh, 0a2h // cpuid
 mov edi, regeax
 mov [edi], eax
 mov edi, regebx
 mov [edi], ebx
 mov edi, regecx
 mov [edi], ecx
 mov edi, regedx
 mov [edi], edx
 popad
end;

function inttohex(w: Word):string;
const
  hCh:array [0..$F] of Char='0123456789ABCDEF';
begin
  inttohex:=hCh[Hi(w)shr 4]+hCh[Hi(w)and $F]+hCh[Lo(w)shr 4]+hCh[Lo(w)and $F];
end;

function getPSNstring:string;
begin
 result :=IntToHex(hiword(psn[0]))+'-'+IntToHex(loword(psn[0]))+'-'
+IntToHex(hiword(psn[1]))+'-'+IntToHex(loword(psn[1]))+'-'
+IntToHex(hiword(psn[2]))+'-'+IntToHex(loword(psn[2]));
end;

procedure GetCPUInfo;
var
 a, b, c, d:UInt32;
 i:Int32;

begin
 if (CPUidSupported) then begin
  baseLevel := $00000000;
  callCPUid(baseLevel, maxLevel, b, c, d);

  SetLength(vendorIdString, 12);
  UInt32(Pointer(@vendorIdString[1])^) := b;
  UInt32(Pointer(@vendorIdString[5])^) := d;
  UInt32(Pointer(@vendorIdString[9])^) := c;

  i := ord(venAMD);
  while (i<ord(high(TCPUVendor))) and (vendorIdStrings[TCPUVendor(i)]<>vendorIdString)
   do inc(i);
  vendor:=TCPUVendor(i);

  callCPUid(baseLevel+1, a, b, c, d);
  if (baseLevel=0) then cpuType := TCPUType((a shr 12) and $3);
  family := TCPUFamily((a shr 8) and $F);
  if (vendor=venIntel) and (family=$F) then family := TCPUFamily((a shr 20) and $FF);
  model := TCPUModel((a shr 4) and $F);
  if (vendor=venIntel) and (model=$F) then model := TCPUModel((a shr 16) and $F);
  stepping := TCPUStepping(a and $F);

  if ((vendor=venAMD) and (family=5) and (model >= 1)) or
   ((vendor in [venAMD, venCyrix, venCentaur]) and (family >= 5)) then begin
   baseLevel := UInt32($80000000);
   callCPUid(baseLevel, maxLevel, b, c, d);
   maxLevel := maxLevel and not UInt32($80000000);
   callCPUid(baseLevel+1, a, b, c, d);
   family := TCPUFamily((a shr 8) and $F);
   model := TCPUModel((a shr 4) and $F);
   stepping := TCPUStepping(a and $F);
  end;

  if (vendor=venIntel) and (model=5) and (maxlevel < 1) then maxlevel := 1;

  if (maxLevel >= 1) then begin
   if (vendor=venIntel) then begin //serial number
    callCPUid(baseLevel+1, psn[0], b, c, d);
    callCPUid(baseLevel+3, a, b, psn[1], psn[2]);
    sn:=#13#10'Serial No.: '+getPSNString;
   end;
  end;
 end else begin
  if (Is_386) then family := 3 else
  if (Is_486) then family := 4 else
  family := 5;
  vendor := venUnknown;
  model := $ff;
  stepping := $ff;
 end;
 clockrate := getCPUClockrate(family);
end;

function GetCPUString:ansiString;
const
 AMDBrandTable:array[4..7, 0..$f] of ansistring=(
  ('', '', '', '80486DX2', '', '', '', '80486DX2 Write back enhanced', '80486DX4', '80486DX4 Write back enhanced', '', '', '', '', '5x86', '5x86 Write back enhanced'),
  ('SSA5 (PR75, PR90, PR100)','5k86 (PR120 or PR133)', '5k86 (PR166)', '5k86 (PR200)', '', '',  'K6 (0.30m)', 'K6 (0.25 um)', 'K6-2', 'K6-III', '', '', '', '', '', ''),
  ('K7','K7', 'Duron', 'K7', '', '', 'K6 (0.30 um)', 'K6 (0.25 um)', ' K6-2', 'K6-III','','','','K6-II+ or K6-III+', '', ''),
  ('','K7', 'K7', 'K7 Duron', 'K7 Thunderbird', '', '', '', '', '', '', '', '', '', '', ''));
 CyrixBrandTable:array[4..6, 2..9] of ansistring=(
  ('', '', '', '', '', '', '', '5x86'),
  ('6x86', '', 'MediaGX, GXm', '', '', '', '', ''),
  ('6x86MX', '', '', '', '', '', '', '')
  );
 IntelBrandTable:array[4..6, 0..8] of ansistring=(
  ('i80486DX', 'i80486DX', 'i80486SX', 'i80486DX2', 'i80486SL', 'i80486SX2', 'i80486DX2 Write back enhanced', 'i80486DX4', 'i804864 Write back enhanced'),
  ('Pentium A-Step', 'Pentium', 'Pentium 54C', 'Pentium 54T OverDrive', 'Pentium 55C', '', '', 'Pentium 54C', 'Pentium 55C (0.25m)'),
  ('Pentium Pro A-Step', 'Pentium Pro', '', 'Pentium II', '', 'Pentium II or Celeron or PII Xeon', 'Celeron A (Mendocino)', 'Pentium III or Pentium III Xeon', 'Pentium III or Pentium III Xeon')
  );

var
 tableString:ansiString;
begin
   tableString := 'unknown';
   case vendor of
    venAMD    :if (family in [4..7]) and (model in [0..$f]) then tableString := AMDBrandTable[family, model];
    venCyrix  :if (family in [4..6]) and (model in [2..9]) then tableString := CyrixBrandTable[family, model];
    venIntel  :if (family=0) then begin
                if model=0 then tableString:='Pentium IV'
               end else begin
                if (family in [4..6]) and (model in [0..8]) then
                 tableString := IntelBrandTable[family, model];
               end;
   end;
 result := CPUVendors[vendor]+' '+tableString;
end;

procedure GetTSCValue(var val:Int64); assembler; pascal;
asm
 pushad
 db 0fh, 31h // rdtsc
 mov edi, VAL
 mov [edi+0], eax
 mov [edi+4], edx
 popad
end;

{ return the current time value }
function GetPITValue(var val:Int64):Bool32; stdcall;
external 'kernel32.dll' name 'QueryPerformanceCounter';

{ measure CPU speed by means of the rdtsc instruction }

function GetTSCSpeed:integer;
const
 MAX_TRIES=20; { number of maximume tries issued }

 TOLERANCE=0.3; { the tolerance in Mhz three subsequent
                        measurement results may differ }
var
 freq:array[0..2] of integer; { storage for the 3 subsequent measurements }
 time0, time1,        { start and endtime (PIT values) }
 stamp0, stamp1,      { start and endcycles (rdtsc values) }
 cycles,
 ticks:Int64;{ used for frequency calculation }
 total:integer;        { total amount of frequency calculated from the 3
                                                        subsequent runs }
 dummy,tries:integer;       { current number of tries }
 oldPriorityClass:integer;

begin
 { initialize variables }
 for dummy := 0 to 2 do
  freq[dummy] := 0;
 tries := 0;
 time0 := 0; time1 := 0; stamp0 := 0; stamp1 := 0;
 repeat
  { increment number of tries }
  tries := tries+1;
  { rotate frequency values in storage }
  freq[2] := freq[1];
  freq[1] := freq[0];
  // maximize priority for most exact results
  oldPriorityClass := GetPriorityClass(GetCurrentProcess);
  SetPriorityClass(GetCurrentProcess, REALTIME_PRIORITY_CLASS);

  { this accounts for overhead later }
  getPITValue(time0);
  time1 := time0;
  while ((time1-time0) < 50) do begin
   getPITvalue(time1);
   getTSCvalue(stamp0);
  end;
  { here's the actual measurement }
  time0 := time1;
  { time for about 10000 cycles }
  while ((time1-time0) < 10000) do begin
   getPITvalue(time1);
   getTSCvalue(stamp1);
  end;

  // restore old priority
  SetPriorityClass(GetCurrentProcess, oldPriorityClass);
  // leave critical section

  cycles := stamp1 - stamp0;

  { calculate frequency }
  ticks := ((time1 - time0) * 1000000) / TIMER_RATE;
  freq[0] := round(cycles / ticks);
  total := freq[0]+freq[1]+freq[2];

  { if we've made enough tries, exit }
  if (tries > MAX_TRIES) then break;
  { loop until three subsequent measurements are within tolerance }
 until ((abs((3*freq[0])-total) <= TOLERANCE) and
  (abs((3*freq[1])-total) <= TOLERANCE) and
  (abs((3*freq[2])-total) <= TOLERANCE));
 { return speed value }
 getTSCspeed := total div 3;
end;

function getCPUclockrate(family:UInt8):integer;
const
 TIMING_RUNS=5;
var
 times,avgspeed:integer;
begin
 avgspeed:=0;
 for times:=1 to TIMING_RUNS do begin
  avgspeed:=avgspeed+getTSCspeed;
 end;
 result := (avgspeed div TIMING_RUNS);
end;

function QueryPerformanceFrequency(var frequency:Int64):Bool32; stdcall;
external 'kernel32.dll' name 'QueryPerformanceFrequency';

function cpu_info:ansistring;
begin
 EnterCriticalSection(critical);
 if resi='' then begin
  QueryPerformanceFrequency(timer_rate);
  sn:='';
  getCPUInfo;
  resi:='CPU info for '+GetCPUString+
  #13#10'Vendor    : '+CPUVendors[vendor]+
  #13#10'Vendor Id : '+vendorIdString+
  #13#10'Type      : '+CPUTypes[cpuType]+
  #13#10'Family    : '+strl(family)+
  #13#10'Model     : '+strl(model)+
  #13#10'Stepping  : '+strl(stepping)+sn+
  #13#10'Clock     : '+strl(clockrate)+' Mhz';
 end;
 LeaveCriticalSection(critical);
 cpu_info:=resi;
end;
