{$ifndef minimal}
{$define mod_config}{$define mod_conf}

const
 defaultlogformat='%r';
type
 perrorpage=^errorpage;
 errorpage=record
  next:perrorpage;
  code:integer;
  page:string;
 end;

 pindex=^index; {directory indexes}
 index=record
  next:pindex;
  name:string;
 end;

 pauth=^auth;
 auth=record
  next:pauth;
  user,pass:string;
  encoding:integer;{0=plaintext,5=md5}
 end;

 piptable=^iptable;
 iptable=record {allow vykousne z ban, ban vykousne z allow. Posledni zaznam vyhrava}
  next:piptable;
  lo_ip,hi_ip:longint;
 end;

 pactdef=^actdef;
 actdef=record
  next:pactdef;
  mask:string;{'*.exe, nebo obecne cokoliv s otaznikem. Bere se jen posleni cast (vlastni jmeno souboru)'}
  mime:string;{""=default}
  act:action{ac_send,ac_forbid,ac_exec,ac_execnph,ac_sendnph};
  param:string;{parametry pro akci, u exec je to jmeno interpreteru (prazdny string znamena poustet primo)}
 end;

 pconfig=^config;
 config=record
  host:string;{Pro co to plati. ''=pro vsechno, '.host.cz' plati pro '*.host.cz'}
  root:string;{Pro co to plati. ''=pro vsechno, co na tohle zacina (automaticky
               se ke konci retezce priplacne lomitko, pokud tam neni)}
  diskroot:string;{cesta k HTML souborum na disku. Taky se k tomu priplacne lomitko}
  accesslog:string;{cesta k logu, ''=neni}
  errorlog:string;{cesta k logu, ''=neni}
  logformat:string;
  dirlist:integer;{directory listing type (0=ban, jinak standard)}
  indexes:pindex;{directory indexes (index.htm, index.php, etc ...)}
  error:perrorpage;{error pages}
  auth:pauth;{nil=bez autorizace}
  ipban:piptable;{nil=zadny ban}
  actdef:pactdef;{nil=bez spec.akci}
  next:pconfig;{nil=posledni}
 end;

 generalconfig=record
  proxyban:piptable;{nil=zadny ban}
  adminban:piptable;{nil=zadny ban}
  adminauth:pauth;{autorizace pro admin stranky}
  logadmin:boolean;{logovat pristupy na monitor?}
  port:word;
  silent:boolean;{nehazet chybove hlasky (messagebox) pri chybe}
  defcfg:pconfig;
  firstcfg:pconfig;
  actcfg:pconfig;
  lastcfg:pconfig;
 end;

var
 gcfg:generalconfig;

procedure genericfree(p:pointer);
var
 q:pointer;
begin
 while p<>nil do begin
  q:=piptable(p)^.next;
  {$ifdef debug}_freemem(p);{$else}freemem(p);{$endif}
  p:=q;
 end;
end;

procedure free_conf(p:pconfig);
begin
 {vsechny maji next jako prvni polozku}
 genericfree(p^.auth);
 genericfree(p^.ipban);
 genericfree(p^.actdef);
 genericfree(p^.error);
 genericfree(p^.indexes);
 freemem(p,sizeof(config));
end;

procedure freeconfig;
var
 p:pconfig;
begin
 with gcfg do begin
  while firstcfg<>nil do begin
   p:=actcfg^.next;
   free_conf(actcfg);
   actcfg:=p;
  end;
  lastcfg:=nil;
  actcfg:=nil;
  free_conf(defcfg);
  defcfg:=nil;
 end;
end;

procedure createdefaultconfig;
begin
 getzero(gcfg.defcfg,sizeof(config));
 gcfg.port:=server_min_port;
 with gcfg.defcfg^ do begin
  host:='';
  root:='';
  diskroot:='';
  accesslog:='';
  errorlog:='';
  logformat:=defaultlogformat;
  dirlist:=1;
  error:=nil;
  auth:=nil;
  ipban:=nil;
  actdef:=nil;
  next:=nil;
  indexes:=nil;
 end;
end;

function dup_indexes(p:pindex):pindex;
var
 r,q:pindex;
begin
 r:=nil;
 dup_indexes:=nil;
 while p<>nil do begin
  getzero(q,sizeof(q^));
  q^.name:=p^.name;
  if r=nil then dup_indexes:=q else r^.next:=q;
  r:=q;
  {$ifdef debug}if p=p^.next then fatal('index circle detected!!!');{$endif}
  p:=p^.next;
 end;
end;

function dup_error(p:perrorpage):perrorpage;
var
 r,q:perrorpage;
begin
 r:=nil;
 dup_error:=nil;
 while p<>nil do begin
  getzero(q,sizeof(q^));
  q^.page:=p^.page;
  q^.code:=p^.code;
  if r=nil then dup_error:=q else r^.next:=q;
  r:=q;
  {$ifdef debug}if p=p^.next then fatal('error circle detected!!!');{$endif}
  p:=p^.next;
 end;
end;

function dup_auth(p:pauth):pauth;
var
 r,q:pauth;
begin
 r:=nil;
 dup_auth:=nil;
 while p<>nil do begin
  getzero(q,sizeof(q^));
  q^.user:=p^.user;
  q^.pass:=p^.pass;
  q^.encoding:=p^.encoding;
  if r=nil then dup_auth:=q else r^.next:=q;
  r:=q;
  {$ifdef debug}if p=p^.next then fatal('auth circle detected!!!');{$endif}
  p:=p^.next;
 end;
end;

function dup_ipban(p:piptable):piptable;
var
 r,q:piptable;
begin
 dup_ipban:=nil;
 r:=nil;
 while p<>nil do begin
  getzero(q,sizeof(q^));
  q^.lo_ip:=p^.lo_ip;
  q^.hi_ip:=p^.hi_ip;
  if r=nil then dup_ipban:=q else r^.next:=q;
  r:=q;
  {$ifdef debug}if p=p^.next then fatal('ipban circle detected!!!');{$endif}
  p:=p^.next;
 end;
end;

function dup_actdef(p:pactdef):pactdef;
var
 r,q:pactdef;
begin
 r:=nil;
 dup_actdef:=nil;
 while p<>nil do begin
  getzero(q,sizeof(q^));
  q^.mask:=p^.mask;
  q^.mime:=p^.mime;
  q^.param:=p^.param;
  q^.act:=p^.act;
  if r=nil then dup_actdef:=q else r^.next:=q;
  r:=q;
  {$ifdef debug}if p=p^.next then fatal('actdef circle detected!!!');{$endif}
  p:=p^.next;
 end;
end;

procedure copyconfig(c_from,c_to:pconfig);
begin
 c_to^.host:=c_from^.host;
 c_to^.root:=c_from^.host;
 c_to^.diskroot:=c_from^.diskroot;
 c_to^.accesslog:=c_from^.accesslog;
 c_to^.errorlog:=c_from^.errorlog;
 c_to^.logformat:=c_from^.logformat;
 c_to^.dirlist:=c_from^.dirlist;
 c_to^.indexes:=dup_indexes(c_from^.indexes);
 c_to^.error:=dup_error(c_from^.error);
 c_to^.auth:=dup_auth(c_from^.auth);
 c_to^.ipban:=dup_ipban(c_from^.ipban);
 c_to^.actdef:=dup_actdef(c_from^.actdef);
 c_to^.next:=nil;
end;

function newconfig:pconfig;
var
 c:pconfig;
begin
 with gcfg do begin
  getzero(c,sizeof(config));
  {$ifdef debug}writeln('CFG:memory allocated at '+strl(integer(c)));{$endif}
  copyconfig(defcfg,c);
  if firstcfg=nil then firstcfg:=c;
  if lastcfg<>nil then lastcfg^.next:=c;
  lastcfg:=c;
  newconfig:=c;
 end;
end;

procedure hostroot(var host,root:string;const src:string);
var
 l,i:integer;
begin
 l:=length(src);
 i:=pos('/',src);
 if i=0 then host:=trim(copy(src,1,l)) else begin
 host:=s_locase(trim(copy(src,1,i-1)));
 root:=s_locase(trim(copy(src,i+1,l-i)));
 l:=length(root);
 if (l>0) then begin
  for i:=1 to l do if root[i]='/' then root[i]:='\';
   if (root[l]<>'\') then root:=root+'\';
  end;
 end;
 {$ifdef debug}writeln('CFG:root='+root+',host='+host);{$endif}
end;

function createalias(src:pconfig;const val:string):pconfig;
var
 c:pconfig;
 newhost,newroot:string;
begin
 hostroot(newhost,newroot,val);
 with gcfg do begin
  getzero(c,sizeof(config));
  copyconfig(src,c);
  c^.host:=newhost;
  c^.root:=newroot;
  if firstcfg=nil then firstcfg:=c;
  if lastcfg<>nil then lastcfg^.next:=c;
  lastcfg:=c;
  createalias:=c;
 end;
end;

procedure add_auth2(var p:pauth;const usr:string;enc:integer;const pass:string);
var
 x:pauth;
begin
 getzero(x,sizeof(auth));
 x^.user:=usr;
 if enc=5 then x^.pass:=fromhex(pass) else x^.pass:=pass;
 x^.encoding:=enc;
 x^.next:=p;
 p:=x;
end;

procedure explode(ex:char;const line:string;var a:array of string;num:integer);
var
 lp,p,l,i:integer;
begin
 l:=length(line);
 if l=0 then exit;
 lp:=1;p:=1;
 for i:=0 to num-1 do begin
  while line[p]<>ex do begin
   inc(p);
   if p>l then break;
  end;
  a[i]:=copy(line,lp,p-lp);
  inc(p);
  lp:=p;
  if p>l then exit;{dalsi nebudou}
 end;
end;

procedure add_auth(var p:pauth;const line:string);
var
 a:array[0..2] of string;
begin
 explode(':',line,a,3);
 add_auth2(p,a[0],ival(a[1],0),a[2]);
end;

procedure set_index(var pp:pindex;const val:string);
var
 lp,p,l:integer;
 q:pindex;
begin
 genericfree(pp);
 pp:=nil;
 l:=length(val);
 if l=0 then exit;
 lp:=l;p:=l;
 repeat
  while val[p]<>',' do begin
   dec(p);
   if p<=0 then break;
  end;
  getzero(q,sizeof(index));
  q^.name:=copy(val,p+1,lp-p);
  q^.next:=pp;
  pp:=q;
  dec(p);
  lp:=p;
  if p<1 then exit;{dalsi nebudou}
 until false;
end;

function conv_error(s:string):string;
var
 i,l:integer;
begin
 l:=length(s);
 if l>0 then begin
  for i:=1 to l do if s[i]='/' then s[i]:='\';
  if s[1]='\' then s:=copy(s,2,l);
 end;
 result:=s;
end;

procedure add_error(var p:perrorpage;const val:string);
var
 a:array[0..1] of string;
 i:integer;
 q:perrorpage;
begin
 explode(';',val,a,2);
 i:=ival(a[0],0);
 if i=0 then exit;{sorry, ale chyba musi byt cislo.}
 q:=p;
 a[1]:=conv_error(a[1]);
 while q<>nil do begin
  if q^.code=i then begin
   q^.page:=a[1];{nahradi se}
   exit;
  end;
  q:=q^.next;
 end;
 getzero(q,sizeof(errorpage));
 q^.next:=p;
 q^.code:=i;
 q^.page:=a[1];
 p:=q;
end;

function parse_action(const s:string):action;
begin
 if s='exec' then result:=ac_exec else
 if s='execnph' then result:=ac_execnph else
 if s='sendnph' then result:=ac_sendnph else
 if s='forbid' then result:=ac_forbid else
 result:=ac_send;
end;

procedure add_action(var p:pactdef;const val:string);
var
 a:array[0..3] of string;
 z:pactdef;
begin
 explode(';',val,a,4);
 getzero(z,sizeof(actdef));
 z^.next:=p;
 z^.mask:=a[0];
 z^.mime:=a[1];
 z^.act:=parse_action(s_locase(a[2]));
 z^.param:=a[3];
 p:=z;
end;

procedure add_ipban2(var p:piptable;start,stop:longint);
var
 x:piptable;
begin
 getzero(x,sizeof(iptable));
 x^.lo_ip:=start;
 x^.hi_ip:=stop;
 x^.next:=p;
 p:=x;
end;

function l(i,j:longint):boolean;
begin
 l:=(i-$80000000)<(j-$80000000);
end;

function le(i,j:longint):boolean;
begin
 le:=(i-$80000000)<=(j-$80000000);
end;

function ge(i,j:longint):boolean;
begin
 ge:=(i-$80000000)>=(j-$80000000);
end;

function ip_in(ip,lo,hi:integer):boolean;
begin
 {ip>=lo and ip<=hi}
 ip_in:=ge(ip,lo) and le(ip,hi);
end;

procedure getstartstop(const s:string;var st,sp:longint);
var
 i,start,stop:longint;
begin
 if s_locase(s)='all' then begin
  start:=0;
  stop:=$ffffffff;
 end else begin
  stop:=pos('-',s);
  if stop=0 then begin
   start:=(resolve(pchar(s)));
   stop:=start;
  end else begin
   start:=(resolve(pchar(copy(s,1,stop-1))));
   stop:=(resolve(pchar(copy(s,stop+1,$7fff))));
  end;
  if l(stop,start) then begin
   i:=start;
   start:=stop;
   stop:=i;
  end;
 end;
 st:=start;
 sp:=stop;
end;

procedure clear_ipban_table(src:piptable;st,sp:longint);
var
 x:piptable;
begin
 x:=src^.next;
 while x<>nil do with x^ do begin
  if le(st,lo_ip) and ge(sp,hi_ip) then begin
  {$ifdef debug}writeln('Killing: '+str_ip(lo_ip)+'..'+str_ip(hi_ip));{$endif}
  {$ifdef debug}writeln('CauseOf: '+str_ip(st)+'..'+str_ip(sp));{$endif}
   src^.next:=x^.next;
   freemem(x,sizeof(iptable));
   x:=src^.next;
  {$ifdef debug}writeln('DoneKill');{$endif}
  end else begin
   src:=x;
   x:=x^.next;
  end;
 end;
end;

procedure add_ipban(var p:piptable;const s:string);
var
 start,stop:longint;
 x:piptable;
begin
 getstartstop(s,start,stop);
 x:=p;
 while x<>nil do with x^ do begin
  if le(lo_ip,start) and ge(hi_ip,start) then begin
   {start obsazen v intervalu}
   if l(hi_ip,stop) then hi_ip:=stop;
   clear_ipban_table(x,lo_ip,stop);
   exit;
  end;
  if le(lo_ip,stop) and ge(hi_ip,stop) then begin
   {stop obsazen v intervalu -> start je mimo}
   lo_ip:=start;
   clear_ipban_table(x,start,hi_ip);
   exit;
  end;
  x:=next;
 end;
 add_ipban2(p,start,stop);
end;

procedure add_ipallow(var p:piptable;const s:string);
var
 start,stop:longint;
 w,x,q:piptable;
begin
 getstartstop(s,start,stop);
 x:=p;
 q:=nil;
 while x<>nil do with x^ do begin
  if le(start,lo_ip) and ge(stop,hi_ip) then begin
   {zrusi se cely interval}
   if q=nil then p:=next else q^.next:=next;
   w:=x;
   x:=next;
   freemem(w,sizeof(iptable));
   continue;
  end;
  {interval se rusit nebude a ma delku min. dva}
  if le(start,lo_ip) then begin
   {zacatek je mimo interval nebo na jeho zacatku}
   if ip_in(stop,lo_ip,hi_ip) {konec v intervalu} then begin
    lo_ip:=stop+1;
   end; {else nic se nedeje}
  end else if le(start,hi_ip) then begin
   if ge(stop,hi_ip) then hi_ip:=start-1 else begin
    {potreba rozdelit na dve casti}
    add_ipban2(p,stop+1,hi_ip);
    hi_ip:=start-1;
   end;
  end;
  {else nic se nedeje}
  q:=x;
  x:=next;
 end;
 {konec vykusovani}
end;

procedure fastconfig;
begin
 createdefaultconfig;
 add_ipban(gcfg.defcfg^.ipban,'127.0.0.1');
 add_auth2(gcfg.adminauth,'ex',0,'server');
end;

procedure{function} loadfile(const s:string;var f:string){:string};
var
 h:thandle;
 siz:u_val;
begin
 h:=createfile(pchar(s),generic_read,file_share_read,nil,
    open_existing,file_flag_sequential_scan,0);
 if h=INVALID_HANDLE_VALUE then exit;
 siz:=getfilesize(h,nil);
 setlength(f,siz);
 if not(readfile(h,f[1],siz,siz,nil)) then setlength(f,0);
 closehandle(h);
end;

function readline(const s:string;var o:string;var pp:integer;l:integer):boolean;
var
 p:integer;
begin
 p:=pp;
 setlength(o,0);
 readline:=true;
 while p<=l do begin
  if not(s[p] in [' ',#9,#13]) then break;
  inc(p);
 end;
 while p<=l do begin
  if s[p]=#10 then begin
   inc(p);
   pp:=p;
   exit;
  end;
  if s[p]='#' then begin
   while p<l do begin
    inc(p);
    if s[p]=#10 then begin
     inc(p);
     break;
    end;
   end;
   pp:=p;
   exit;
  end;
  if s[p]<>#13 then o:=o+s[p];
  inc(p);
 end;
 pp:=p;
 readline:=(length(o)>0);
end;

const
 crlf=#10;

function loadconfig(const s:string):string;
var
 c:pconfig;
 warn,f,key,value,line:string;
 l,i,len,p:integer;
 aliasline:boolean;
begin
 aliasline:=false;
 f:='';
 warn:='';
 createdefaultconfig;
 {$ifdef debug}writeln('CFG:default created');{$endif}
 if length(s)=0 then begin
  {no config}
  result:='';
  exit;
 end;
 loadfile(s,f);
 len:=length(f);
 if len=0 then begin
  loadconfig:='Config file empty or invalid'+crlf;
  exit;
 end;
 with gcfg do begin
  c:=defcfg;
  p:=1;
  while readline(f,line,p,len) do begin
   line:=trim(line);
   l:=length(line);
   if (l>0) then begin
    if (line[1]='<') and (line[l]='>') then begin
     if s_locase(line)='</root>' then begin
      c:=defcfg;
      aliasline:=false;
     end else if beginwithi(line,'<root ') then begin
      {$ifdef debug}writeln('CFG:new root: '+line);{$endif}
      aliasline:=false;
      c:=newconfig;
      hostroot(c^.host,c^.root,copy(line,7,l-7));
     end else warn:=warn+('Unrecognized line in config: "'+line+'"'+crlf);
    end else begin
     i:=pos('=',line);
     if i=0 then warn:=warn+('Unrecognized line in config: "'+line+'"'+crlf) else begin
      value:=trim(copy(line,i+1,$7fffffff));
      key:=s_locase(trim(copy(line,1,i-1)));
      if aliasline and (key<>'alias') then warn:=warn+('Warning: line after alias: "'+line+'"'+crlf);
      if key='alias' then begin
       createalias(c,value);
       aliasline:=true;
      end else if key='diskroot' then begin
       l:=length(value);
       if (l>0) then begin
        for i:=1 to l do if value[i]='/' then value[i]:='\';
        if (value[l]<>'\') then value:=value+'\';
       end;
       c^.diskroot:=value;
      end else if key='logformat' then begin
       c^.logformat:=value;
      end else if key='accesslog' then begin
       if value='nul' then value:='';
       c^.accesslog:=value;
      end else if key='errorlog' then begin
       if value='nul' then value:='';
       c^.errorlog:=value;
      end else if key='logmonitor' then begin
       gcfg.logadmin:=bval(value);
      end else if key='port' then begin
       gcfg.port:=ival(value,server_min_port);
      end else if key='silent' then begin
       gcfg.silent:=bval(value);
      end else if key='dirlist' then begin
       c^.dirlist:=ival(value,0);
      end else if key='auth' then begin
       add_auth(c^.auth,value);
      end else if key='monitorauth' then begin
       add_auth(gcfg.adminauth,value);
      end else if key='ipban' then begin
       add_ipban(c^.ipban,value);
      end else if key='ipallow' then begin
       add_ipallow(c^.ipban,value);
      end else if key='proxyban' then begin
       add_ipban(gcfg.proxyban,value);
      end else if key='proxyallow' then begin
       add_ipallow(gcfg.proxyban,value);
      end else if key='monitorban' then begin
       add_ipban(gcfg.adminban,value);
      end else if key='monitorallow' then begin
       add_ipallow(gcfg.adminban,value);
      end else if key='index' then begin
       set_index(c^.indexes,value);
      end else if key='action' then begin
       add_action(c^.actdef,value);
      end else if key='error' then begin
       add_error(c^.error,value);
      end else begin
       warn:=warn+('Unrecognized line in config: "'+line+'"'+crlf);
      end;
     end;
    end;
   end;
  end;
 end;
 {$ifdef debug}writeln('CFG:config loaded');{$endif}
 server_port:=gcfg.port;
 server_min_port:=gcfg.port;
 server_max_port:=gcfg.port;
 loadconfig:=warn;
end;

function endwithi2(const a,b:string):boolean;
var
 i,la,lb:integer;
begin
 lb:=length(b);
 la:=length(a);
 if lb<2 then result:=true else
 if la<lb then result:=false else begin
  la:=la-lb;
  for i:=2 to lb do begin
   if a[i+la]<>b[i] then begin
    result:=false;
    exit;
   end;
  end;
  result:=true;
 end;
end;

function matchdomain(const a,b:string):boolean;
begin
 if length(b)=0 then result:=true else begin
  if b[1]='.' then result:=endwithi2(a,b) else result:=(a=b);
 end;
end;

function askconfig(var ss:send_struct):pconfig;
var
 x:pconfig;
 q:ansistring;
begin
 x:=gcfg.firstcfg;
 q:=ss.filename+'\';
 while x<>nil do with x^ do begin
  if not(matchdomain(ss.host,host)) then x:=next
  else if not(beginwith(q,root)) then x:=next else break;
 end;
 if x=nil then x:=gcfg.defcfg;
 askconfig:=x;
end;

function banned(x:piptable;ip:integer):boolean;
begin
 while x<>nil do with x^ do begin
  if le(lo_ip,ip) and le(ip,hi_ip) then begin
  {$ifdef debug}writeln('IP banned: '+str_ip(ip)+' in '+str_ip(lo_ip)+'..'+str_ip(hi_ip));{$endif}
   result:=true;
   exit;
  end;
  x:=next;
 end;
 result:=false;
end;

{--logfiles}

type
 phashtable=^hashtable;
 hashtable=record
  s:string;
  i:thandle;
  next:phashtable;
 end;

procedure ht_set(var q:phashtable;s:string;value:integer);
var
 z:phashtable;
begin
 s:=s_locase(s);
 z:=q;
 while z<>nil do begin
  if z^.s=s then begin
   z^.i:=value;
   exit;
  end;
  z:=z^.next;
 end;
 getzero(z,sizeof(hashtable));
 z^.s:=s;
 z^.i:=value;
 z^.next:=q;
 q:=z;
end;

function ht_get(q:phashtable;s:string;default_val:integer):integer;
begin
 s:=s_locase(s);
 while q<>nil do begin
  if q^.s=s then begin
   ht_get:=q^.i;
   exit;
  end;
  q:=q^.next;
 end;
 ht_get:=default_val;
end;

procedure ht_del(var q:phashtable;s:string);
var
 z,w:phashtable;
begin
 s:=s_locase(s);
 w:=q;
 z:=nil;
 while w<>nil do begin
  if w^.s=s then begin
   if z=nil then begin
    q:=w^.next;
   end else begin
    z^.next:=w^.next;
   end;
   freemem(w,sizeof(hashtable));
  end;
  z:=w;
  w:=w^.next;
 end;
 {not found,exiting}
end;

{$define log_mutex}
{todo: s tema mutexama byly na w2k problemy}

var
 hlog{$ifdef log_mutex},hmut{$endif}:phashtable;

function open_log(const s:string):thandle;
var
 t{$ifdef log_mutex},q{$endif}:thandle;
 d:integer;
begin
 {$ifdef log_mutex}
 q:=ht_get(hmut,s,invalid_handle_value);
 if q=invalid_handle_value then begin
  q:=createmutex(nil,false,pchar(s));
  ht_set(hmut,s,q);
 end;
 waitforsingleobject(q,infinite);
 {$endif}
 t:=ht_get(hlog,s,invalid_handle_value);
 if t=invalid_handle_value then begin
  t:=createfile(pchar(s),GENERIC_WRITE,FILE_SHARE_READ,nil,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
  d:=0;
  setfilepointer(t,0,@d,file_end);
  ht_set(hlog,s,t);
 end;
 {$ifdef log_mutex}
 releasemutex(q);
 {$endif}
 open_log:=t;
end;

procedure close_log(const s:string);
var
 t:thandle;
begin
 t:=ht_get(hlog,s,invalid_handle_value);
 closehandle(t);
 ht_del(hlog,s);
 {$ifdef log_mutex}
 t:=ht_get(hmut,s,invalid_handle_value);
 closehandle(t);
 ht_del(hmut,s);
 {$endif}
end;

procedure writeline(h:thandle;s:string);
var
 dummy:integer;
begin
 s:=s+#10;
 writefile(h,s[1],length(s),dummy,nil);
end;

procedure write_logline(const logfile,logline:string);
var
 h:thandle;
begin
 h:=open_log(logfile);
 writeline(h,logline);
end;

function nowtime:string;
var s:tsystemtime;
begin
 getsystemtime(s);
 result:=
strl0(s.wday,2)+'/'+
strl0(s.wmonth,2)+'/'+
strl0(s.wyear,4)+':'+
strl0(s.whour,2)+':'+
strl0(s.wminute,2)+':'+
strl0(s.wsecond,2);
end;

function format(var ss:send_struct;const fmt:string):string;
var
 what,en:string;
 i,l:integer;
 ispct:boolean;
begin
 en:='';
 l:=length(fmt);
 i:=1;
 ispct:=false;
 while i<=l do begin
  if ispct then begin
   case fmt[i] of
    '%':en:=en+'%';
    '{':begin
         what:='';
         while i<=l do begin
          inc(i);
          if fmt[i]='}' then break else what:=what+fmt[i];
         end;
         en:=en+getheader(ss,what);
        end;
    'a':en:=en+str_ip(ss.remote);
    'B':begin
         if ss.contentlen>0 then en:=en+strl(ss.contentlen)
         else en:=en+'-';
        end;
    'm':en:=en+ss.method;
    'q':en:=en+ss.qs;
    'r':en:=en+ss.line1;
    's':en:=en+strl(ss.errorcode);
    't':en:=en+nowtime;
    'u':if ss.verified then en:=en+ss.verified_user else en:=en+'-';
    'U':en:=en+ss.url;
    'v':en:=en+pconfig(ss.config)^.root;
   else en:=en+'??%'+fmt[i]+'??';
   end;
   ispct:=false;
  end else begin
   if fmt[i]='%' then ispct:=true else en:=en+fmt[i];
  end;
  inc(i);
 end;
 result:=en;
end;

procedure write_fatal_elog(const e:string);
var
 P:pconfig;
begin
 p:=gcfg.defcfg;
 if length(p^.errorlog)=0 then exit;
 write_logline(p^.errorlog,nowtime+' '+e);
end;

procedure write_elog(var ss:send_struct;const e:string);
var
 P:pconfig;
begin
 p:=ss.config;
 if length(p^.errorlog)=0 then exit;
 write_logline(p^.errorlog,format(ss,'%t %a ')+e);
end;

procedure write_alog(var ss:send_struct);
var
 P:pconfig;
begin
 if ss.privileged and not(gcfg.logadmin) then exit;
 p:=ss.config;
 if length(p^.accesslog)=0 then exit;
 write_logline(p^.accesslog,format(ss,p^.logformat));
end;

{$endif}